/* slice.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/* 2-D slice module. */



#include <stdio.h>
#include <math.h>
#include "globals.h"
#include "graphics.h"
#include "matrix.h"
#include "queue.h"


#ifndef TRUE
#   define TRUE 1
#endif

#ifndef FALSE
#   define FALSE 0
#endif




/*** clamp ************************************************************
   Clamp a value 'val' to the interval [min,max].
   Input:  min - the min value of the interval.
           val - the value to clamp.
           max - the max value of the interval.
   Returned:  the clamped value.
**********************************************************************/ 
static float clamp( min, val, max )
float min,val,max;
{
   if (val<min)  return min;
   if (val>max)  return max;
   return val;
}



static void swap( a, b )
float *a, *b;
{
   float t;
   t = *a;  *a = *b;  *b = t;
}


static float sgn( a )
float a;
{
   if (a<0.0)  return -1.0;
   if (a>0.0)  return 1.0;
   return 0.0;
}



/* This array contains projected, normalized box coordinates: */
static float ndcbox[8][3];



/*** calc_ndc_box *****************************************************
   Transform the vertices of the box from graphics coords to ndc.
**********************************************************************/
static void calc_ndc_box()
{
   int i;

   /* Transform view box corners to 2-D normalized coords. */
   ndcbox[0][0] = Xmax;  ndcbox[0][1] = Ymax;  ndcbox[0][2] = Zmin;
   ndcbox[1][0] = Xmax;  ndcbox[1][1] = Ymin;  ndcbox[1][2] = Zmin;
   ndcbox[2][0] = Xmin;  ndcbox[2][1] = Ymin;  ndcbox[2][2] = Zmin;
   ndcbox[3][0] = Xmin;  ndcbox[3][1] = Ymax;  ndcbox[3][2] = Zmin;
   ndcbox[4][0] = Xmax;  ndcbox[4][1] = Ymax;  ndcbox[4][2] = Zmax;
   ndcbox[5][0] = Xmax;  ndcbox[5][1] = Ymin;  ndcbox[5][2] = Zmax;
   ndcbox[6][0] = Xmin;  ndcbox[6][1] = Ymin;  ndcbox[6][2] = Zmax;
   ndcbox[7][0] = Xmin;  ndcbox[7][1] = Ymax;  ndcbox[7][2] = Zmax;
   for (i=0;i<8;i++)
      mat_vecmul( ndcbox[i], CTM );
}




/*** distance_to_hslice ***********************************************
   Calculate the distance from the cursor to the nearest corner of
   the given hslice.
   Input:  curx, cuy - cursor coords in pixels
           level - position of slice in grid coords [0..Nl-1].
   Output:  corner - number of nearest corner in [0..3]
   Returned:  distance from cursor to corner.
**********************************************************************/
static float distance_to_hslice( curx, cury, level, corner )
int curx, cury;
float level;
int *corner;
{
   float cx, cy;
   float t, dist, neardist, p[3];
   int i;

   neardist = 1000000.0;  /* any large value */

   /* Normalize cursor position to [-1,1],[-1,1] */
   cx = ( (float) curx * 2.0 / Width - 1.0 ) * AspectRatio;
   cy = 1.0 - (float) cury * 2.0 / Height;

   /* normalize level to [0,1] */
   t = level / (float) (Nl-1);

   /* Find distance to each corner */
   for (i=0;i<4;i++) {
      p[0] = ndcbox[i+4][0] + t * (ndcbox[i][0] - ndcbox[i+4][0]);
      p[1] = ndcbox[i+4][1] + t * (ndcbox[i][1] - ndcbox[i+4][1]);

      dist = (p[0]-cx)*(p[0]-cx) + (p[1]-cy)*(p[1]-cy);
      if (dist<=neardist) {
	 neardist = dist;
	 *corner = i;
      }
   }

   return neardist;
}



/*** distance_to_vslice ***********************************************
   Calculate the distance from the cursor to the nearest corner of
   the given vslice.
   Input:  curx, cuy - cursor coords in pixels
           r1,c1,r2,c2 - position of slice in grid coords
   Output:  corner - number of nearest corner or edge in 0..5
   Returned:  distance from cursor to corner.
**********************************************************************/
static float distance_to_vslice( curx, cury, r1,c1, r2,c2, corner )
int curx, cury;
float r1,c1, r2,c2;
int *corner;
{
   float cx, cy;
   float r, c, l;
   float t, dist, neardist, p[3];
   int i, k;

   neardist = 1000000.0;  /* any large value */

   /* Normalize cursor position to [-1,1],[-1,1] */
   cx = ( (float) curx * 2.0 / Width - 1.0 ) * AspectRatio;
   cy = 1.0 - (float) cury * 2.0 / Height;

   /* normalize row, column positions to [0,1] */
   r1 = r1 / (float) (Nr-1);
   c1 = c1 / (float) (Nc-1);
   r2 = r2 / (float) (Nr-1);
   c2 = c2 / (float) (Nc-1);

   /* find nearest corner */
   for (i=0;i<4;i++) {
      switch (i) {
	  case 0:  r = r1;  c = c1;  k = 0;  break;
	  case 1:  r = r2;  c = c2;  k = 0;  break;
	  case 2:  r = r1;  c = c1;  k = 4;  break;
	  case 3:  r = r2;  c = c2;  k = 4;  break;
      }
      /* convert r,c,k to ndc x,y,z */
      if (r==0.0) {
	 /* north edge */
	 p[0] = ndcbox[k+3][0] + c * (ndcbox[k+0][0]-ndcbox[k+3][0]);
	 p[1] = ndcbox[k+3][1] + c * (ndcbox[k+0][1]-ndcbox[k+3][1]);
      }
      else if (r==1.0) {
	 /* south edge */
	 p[0] = ndcbox[k+2][0] + c * (ndcbox[k+1][0]-ndcbox[k+2][0]);
	 p[1] = ndcbox[k+2][1] + c * (ndcbox[k+1][1]-ndcbox[k+2][1]);
      }
      else if (c==0.0) {
	 /* west edge */
	 p[0] = ndcbox[k+3][0] + r * (ndcbox[k+2][0]-ndcbox[k+3][0]);
	 p[1] = ndcbox[k+3][1] + r * (ndcbox[k+2][1]-ndcbox[k+3][1]);
      }
      else {
	 /* east edge */
	 p[0] = ndcbox[k+0][0] + r * (ndcbox[k+1][0]-ndcbox[k+0][0]);
	 p[1] = ndcbox[k+0][1] + r * (ndcbox[k+1][1]-ndcbox[k+0][1]);
      }

      dist = (p[0]-cx)*(p[0]-cx) + (p[1]-cy)*(p[1]-cy);
      if (dist<neardist) {
	 neardist = dist;
	 *corner = i;
      }
   }

   /* try to find a closer edge midpoint */
   for (i=4;i<6;i++) {
      c = ( c1 + c2 ) / 2.0;
      r = ( r1 + r2 ) / 2.0;
      l = (i==4) ? 1.0 : 0.0;

      p[0] = ndcbox[7][0] + c*(ndcbox[4][0]-ndcbox[7][0])
	                  + r*(ndcbox[6][0]-ndcbox[7][0])
                          + l*(ndcbox[3][0]-ndcbox[7][0]);
      p[1] = ndcbox[7][1] + c*(ndcbox[4][1]-ndcbox[7][1])
	                  + r*(ndcbox[6][1]-ndcbox[7][1])
                          + l*(ndcbox[3][1]-ndcbox[7][1]);

      dist = (p[0]-cx)*(p[0]-cx) + (p[1]-cy)*(p[1]-cy);
      if (dist<neardist) {
	 neardist = dist;
	 *corner = i;
      }
   }

   return neardist;
}






/*** find_nearest_slice *********************************************
   Given a cursor position, find the slice to be selected for moving.
   Input:  curx,cury - the cursor position in pixels (0,0) = upper-left
           contype   - pointer to int
           contour   - pointer to int
           corner    - pointer to int
   Output:  contype - either HSLICE, VSLICE, CHSLICE, or CVSLICE.
            contour - the number of the nearest contour
            corner - the number of the nearest corner in [0..3] or
                     if a vslice edge midpoint has been selected 4, or 5
   Returned:  TRUE - a contour has been identified.
              FALSE - a contour was not identified.
**********************************************************************/
static int find_nearest_slice( curx, cury, contype, contour, corner )
int curx, cury;
int *contype, *contour, *corner;
{
   float dist, neardist;
   int ip, corn, w;

   calc_ndc_box();

   neardist = 0.06 * 0.06;
   *contype = *contour = *corner = -1;

   /* try horizontal contour slices */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayHSlice[ip]) {
	 dist = distance_to_hslice( curx, cury, HSliceLevel[ip], &corn );
	 if (dist<neardist) {
	    neardist = dist;
	    *contype = HSLICE;
	    *contour = ip;
	    *corner = corn;
	 }
      }
   }


   /* try horizontal color slices */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayCHSlice[ip]) {
	 dist = distance_to_hslice( curx, cury, CHSliceLevel[ip], &corn );
	 if (dist<neardist) {
	    neardist = dist;
	    *contype = CHSLICE;
	    *contour = ip;
	    *corner = corn;
	 }
      }
   }

   /* try vertical contour slices */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayVSlice[ip]) {
	 dist = distance_to_vslice( curx, cury,
				   VSliceR1[ip], VSliceC1[ip],
				   VSliceR2[ip], VSliceC2[ip], &corn );
	 if (dist<neardist) {
	    neardist = dist;
	    *contype = VSLICE;
	    *contour = ip;
	    *corner = corn;
	 }
      }
   }

   /* try vertical colored slices */
   for (ip=0;ip<NumParms+NumClones;ip++) {
      if (DisplayCVSlice[ip]) {
	 dist = distance_to_vslice( curx, cury,
				   CVSliceR1[ip], CVSliceC1[ip],
				   CVSliceR2[ip], CVSliceC2[ip], &corn );
	 if (dist<neardist) {
	    neardist = dist;
	    *contype = CVSLICE;
	    *contour = ip;
	    *corner = corn;
	 }
      }
   }

   /* try horizontal wind slices */
   for (w=0;w<WINDSLICES;w++) {
      if (DisplayHWind[w]) {
	 dist = distance_to_hslice( curx, cury, HWindLevel[w], &corn );
	 if (dist<neardist) {
	    neardist = dist;
	    *contype = HWIND;
	    *contour = w;
	    *corner = corn;
	 }
      }
   }

   /* try vertical wind slice */
   for (w=0;w<WINDSLICES;w++) {
      if (DisplayVWind[w]) {
	 dist = distance_to_vslice( curx, cury, VWindR1[w], VWindC1[w],
				 VWindR2[w],VWindC2[w], &corn );
	 if (dist<neardist) {
	    neardist = dist;
	    *contype = VWIND;
	    *contour = w;
	    *corner = corn;
	 }
      }
   }

   /** ALL DONE!  contype, contour, corner have the results */
   if (*contype!=-1) {
/*      printf("grab dist=%f type %d, ip %d, corner %d\n",
	     neardist, *contype, *contour, *corner ); */
      return TRUE;
   }
   else
      return FALSE;
}




/*** move_hslice ******************************************************
   Move a horizontal slice according to the cursor position.
   Input:  curx, cury - cursor position in pixels
           level - pointer to current slice level
           corner - which corner is 'grabbed'
   Output:  level - modified to change position.
   Returned:  0 = no movement
              1 = slice was moved.
**********************************************************************/
static int move_hslice( curx, cury, level, corner )
int curx, cury;
float *level;
int corner;
{
   float cx,cy;
   float a[3], b[3];
   float dot, mg, comp, lev;

   /* Normalize cursor position to [-1,1]. */
   cx = ( (float) curx * 2.0 / Width - 1.0 ) * AspectRatio;
   cy = 1.0 - (float) cury * 2.0 / Height;

   b[0] = ndcbox[corner+4][0] - ndcbox[corner][0];
   b[1] = ndcbox[corner+4][1] - ndcbox[corner][1];
   a[0] = cx - ndcbox[corner][0];
   a[1] = cy - ndcbox[corner][1];

   /* find component of a on b */
   dot = a[0]*b[0] + a[1]*b[1];
   mg = b[0]*b[0] + b[1]*b[1];
   if (mg>0.0) {
      comp = dot / mg;
      comp = clamp( 0.0, comp, 1.0 );
      *level = (1.0-comp)*(Nl-1);
      return 1;
   }
   return 0;
}



/*** move_vslice ******************************************************
   Move a vertical slice according to the cursor position.
   Input:  curx, cury - cursor position in pixels
           r1,c1,r2,c2 - pointers to current slice position values
           corner - which corner is 'grabbed'
   Output:  r1,c1,r2,c2 - modified to change position.
   Returned:  0 = no movement
              1 = slice was moved.
**********************************************************************/
static int move_vslice( curx, cury, r1,c1, r2,c2, corner )
int curx, cury;
float *r1,*c1, *r2,*c2;
int corner;
{
   int kk, i, n;
   float edge[4][2];
   float mag[4], a[3], b[3], bx[3], by[3], d[4], dd[4];
   float comp, dot, mg, lev, neardist;
   float x, y, x1, y1, x2, y2, ax, ay, t;
   float newcol, newrow;
   float cx, cy;
   float sx = (Xmax-Xmin) / (Nc-1), tx = Xmin;
   float sy = -(Ymax-Ymin) / (Nr-1), ty = Ymax;
   float sz = -(Zmax-Zmin) / (Nl-1), tz = Zmax;

   /* Normalize cursor position to [-1,1]. */
   cx = ( (float) curx * 2.0 / Width - 1.0 ) * AspectRatio;
   cy = 1.0 - (float) cury * 2.0 / Height;

   if (corner<4) {
      /* Move a vslice by a corner */
      kk = (corner>=2) ? 4 : 0;

      /* calc top edge vectors (bottom edge vectors are the same.) */
      for (i=0;i<4;i++) {
         edge[i][0] = ndcbox[(i+1)%4][0] - ndcbox[i][0];
         edge[i][1] = ndcbox[(i+1)%4][1] - ndcbox[i][1];
         mag[i] = sqrt( edge[i][0]*edge[i][0] + edge[i][1]*edge[i][1] );
         if (mag[i]==0.)  return;
      }

      /* find the distance from the cursor to each ndcbox corner in 2-D */
      for (i=0;i<4;i++)
         d[i] = sqrt( (cx-ndcbox[i+kk][0])*(cx-ndcbox[i+kk][0])
                    + (cy-ndcbox[i+kk][1])*(cy-ndcbox[i+kk][1]) );

      /* determine which edge of the ndcbox the cursor is nearest */
      dd[0] = (d[0] + d[1]) / mag[0];
      dd[1] = (d[1] + d[2]) / mag[1];
      dd[2] = (d[2] + d[3]) / mag[2];
      dd[3] = (d[3] + d[0]) / mag[3];
      neardist = 10000.0;
      for (i=0; i<4; i++) {
         if (dd[i]<neardist) {
            neardist = dd[i];
            n = i;
         }
      }
      /* 'n' is the nearest edge in [0..3] */

      /* calculate component of a on edge[n] */
      a[0] = cx - ndcbox[n+kk][0];
      a[1] = cy - ndcbox[n+kk][1];
      comp = (a[0]*edge[n][0]+a[1]*edge[n][1]) / mag[n] / mag[n];
      comp = clamp( 0.0, comp, 1.0);

      /* update the X-Section position array */
      switch (n) {
      case 0:
         newcol = Nc-1;
         newrow = (Ymax-comp*(Ymax-Ymin) - ty) / sy;
         break;
      case 1:
         newcol = (Xmax-comp*(Xmax-Xmin) - tx) / sx;
         newrow = Nr-1;
         break;
      case 2:
         newcol = 0;
         newrow = (Ymin+comp*(Ymax-Ymin) - ty) / sy;
         break;
      case 3:
         newcol = (Xmin+comp*(Xmax-Xmin) - tx) / sx;
         newrow = 0;
         break;
      }

      if (corner%2==0) {
	 *r1 = newrow;
	 *c1 = newcol;
      }
      else {
	 *r2 = newrow;
	 *c2 = newcol;
      }
      return 1;

   }
   else {
      /* Move a vslice by its edge. */
      /* calculate X,Y coordinate of cursor in ndcbox coordinates in */
      /* a plane coincident with either the top or bottom of the ndcbox. */
      if (corner==4) {
         /* top plane */
         b[0] = cx - ndcbox[2][0];
         b[1] = cy - ndcbox[2][1];
         bx[0] = ndcbox[1][0] - ndcbox[2][0];
         bx[1] = ndcbox[1][1] - ndcbox[2][1];
         by[0] = ndcbox[3][0] - ndcbox[2][0];
         by[1] = ndcbox[3][1] - ndcbox[2][1];
      }
      else {
         /* bottom plane */
         b[0] = cx - ndcbox[6][0];
         b[1] = cy - ndcbox[6][1];
         bx[0] = ndcbox[5][0] - ndcbox[6][0];
         bx[1] = ndcbox[5][1] - ndcbox[6][1];
         by[0] = ndcbox[7][0] - ndcbox[6][0];
         by[1] = ndcbox[7][1] - ndcbox[6][1];
      }
      /* find components (x and y) of b along bx and by
       * by solving a system of 2 equations and 2 unknowns:
       *     bx[0]*x + by[0]*y = b[0]
       *     bx[1]*x + by[1]*y = b[1]       */
      if (bx[0]==0.0)  return;
      t = by[1] - by[0]*bx[1]/bx[0];
      if (t==0.0)  return;
      y = ( b[1] - b[0]*bx[1]/bx[0] ) / t;
      x = ( b[0] - by[0]*y) / bx[0];

      /* find x, y of cursor in graphics coords */
      x = x * (Xmax-Xmin) + Xmin;
      y = y * (Ymax-Ymin) + Ymin;

      /* a = vector parallel to X-section. */
      ax = ((*c2)*sx+tx) - ((*c1)*sx+tx);
      ay = ((*r2)*sy+ty) - ((*r1)*sy+ty);

      /* find intersection of (x,y) + a and (x,y) - a with the */
      /* sides of the ndcbox (these are the new X-Section coords) */
      if (ax==0.0) {
         /* trivial case */
	 *c1 = *c2 = (clamp(Xmin,x,Xmax)-tx)/sx;
      }
      else if (ay==0.0) {
         /* trivial case */
	 *r1 = *r2 = (clamp(Ymin,y,Ymax)-ty)/sy;
      }
      else {
         /* general intersection case */
         x1 = x + ax * (Ymin - y) / ay;
         y1 = y + ay * (Xmin - x) / ax;
         x2 = x + ax * (Ymax - y) / ay;
         y2 = y + ay * (Xmax - x) / ax;

         /* correct the orientation of the vslice */
         if (ax*ay<0.0)  swap( &x1, &x2 );
         if (sgn(x2-x1)!=sgn(ax)) {
            swap( &x1, &x2 );
            swap( &y1, &y2 );
	 }

         /* Clamp and assign vslice corner coordinates. */
	 *c1 = clamp(0.0, (x1-tx)/sx, (float)(Nc-1));
	 *c2 = clamp(0.0, (x2-tx)/sx, (float)(Nc-1));
	 *r1 = clamp(0.0, (y1-ty)/sy, (float)(Nr-1));
	 *r2 = clamp(0.0, (y2-ty)/sy, (float)(Nr-1));
      }
      return 1;
   }

}






/*** move_slice ****************************************************
   Move the position of a contour.  This function is called to
   respond to a user action.
   Input:  curx, cury - the cursor position in pixels (0,0) = upper-left.
           contype    - the type of contour to move
           ip         - parameter of slice
           corner     - the corner of the contour to use as reference.
	   recalc     - non-zero if slice should be recalculated now.
**********************************************************************/
static void move_slice( curx, cury, contype, ip, corner, recalc )
int curx,cury;
int contype, ip, corner;
int recalc;
{

   if (contype==HSLICE) {
      if (move_hslice( curx, cury, &HSliceLevel[ip], corner ) && recalc) {
	 request_hslice( CurTime, ip );
      }
   }
   else if (contype==CHSLICE) {
      if (move_hslice( curx, cury, &CHSliceLevel[ip], corner ) && recalc) {
	 request_chslice( CurTime, ip );
      }
   }
   else if (contype==VSLICE) {
      if (move_vslice( curx, cury, &VSliceR1[ip], &VSliceC1[ip],
		       &VSliceR2[ip], &VSliceC2[ip], corner ) && recalc) {
	 request_vslice( CurTime, ip );
      }
   }
   else if (contype==CVSLICE) {
      if (move_vslice( curx, cury, &CVSliceR1[ip], &CVSliceC1[ip],
		       &CVSliceR2[ip], &CVSliceC2[ip], corner ) && recalc) {
	 request_cvslice( CurTime, ip );
      }
   }
   else if (contype==HWIND) {
      if (move_hslice( curx, cury, &HWindLevel[ip], corner ) && recalc) {
	 request_hwindslice( CurTime, ip );
      }
   }
   else if (contype==VWIND) {
      if (move_vslice( curx, cury, &VWindR1[ip], &VWindC1[ip],
		       &VWindR2[ip], &VWindC2[ip], corner ) && recalc) {
	 request_vwindslice( CurTime, ip );
      }
   }
   Redraw++;
}




/*** slice_event ****************************************************
   This function is called whenever an input event occurs in the
   main viewing window while in 'slice moving mode'.  If the event
   is used to move a slice TRUE will be returned, else FALSE.
**********************************************************************/
int slice_event( event )
XEvent event;
{
   int result;           /* value returned */
   static int flag;      /* has a contour been selected? */
   static int contype;    /* the type of contour selected */
   static int contour;    /* the contour selectd */
   static int corner;     /* the contour corner selected */
   int recalc;

   result = FALSE;

   if (event.type==ButtonPress && event.xbutton.button==Button3) {
      /* The right mouse button has been pressed. */
      /* Determine which 2-D contour or X-Section corner is */
      /* closest to the current cursor position. */
      flag = find_nearest_slice( event.xbutton.x, event.xbutton.y,
				&contype, &contour, &corner );
      result = TRUE;
   }

   if ( (event.type==MotionNotify || event.type==ButtonPress)
       && flag==TRUE) {
      /* The cursor has been moved while the right mouse button */
      /* is pressed.  Update the position of the contour. */
#ifdef SINGLE_TASK
      recalc = 0;
#else
      recalc = 1;
#endif
      move_slice( event.xbutton.x, event.xbutton.y,
		 contype, contour, corner, recalc );
      result = TRUE;
   }

   if (event.type==ButtonRelease && event.xbutton.button==Button3) {
      /* The right mouse button has been released. */
      move_slice( event.xbutton.x, event.xbutton.y,
		 contype, contour, corner, 1 );
      flag = FALSE;
      result = TRUE;
   }

   return result;
}

